@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 """
-    flaskext.sqlalchemy
-    ~~~~~~~~~~~~~~~~~~~
+    flask_sqlalchemy
+    ~~~~~~~~~~~~~~~~
 
     Adds basic SQLAlchemy support to your application.
 
@@ -12,7 +12,6 @@ from __future__ import absolute_import
 
 import functools
 import os
-import re
 import sys
 import time
 import warnings
@@ -21,18 +20,19 @@ from operator import itemgetter
 from threading import Lock
 
 import sqlalchemy
-from flask import _app_ctx_stack, abort, current_app, g, request
+from flask import _app_ctx_stack, abort, current_app, request
 from flask.signals import Namespace
 from sqlalchemy import event, inspect, orm
 from sqlalchemy.engine.url import make_url
-from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base, \
-    declared_attr
+from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
 from sqlalchemy.orm.exc import UnmappedClassError
 from sqlalchemy.orm.session import Session as SessionBase
 
-from ._compat import iteritems, itervalues, string_types, xrange
+from flask_sqlalchemy.model import Model
+from ._compat import itervalues, string_types, to_str, xrange
+from .model import DefaultMeta
 
-__version__ = '2.2'
+__version__ = '2.3.2'
 
 # the best timer function for the platform
 if sys.platform == 'win32':
@@ -40,8 +40,6 @@ if sys.platform == 'win32':
 else:
     _timer = time.time
 
-_camelcase_re = re.compile(r'([A-Z]+)(?=[a-z0-9])')
-
 _signals = Namespace()
 models_committed = _signals.signal('models-committed')
 before_models_committed = _signals.signal('before-models-committed')
@@ -257,10 +255,12 @@ class _EngineDebuggingSignalEvents(object):
         self, conn, cursor, statement, parameters, context, executemany
     ):
         if current_app:
-            if 'sqlalchemy_queries' not in g:
-                g.sqlalchemy_queries = []
+            try:
+                queries = _app_ctx_stack.top.sqlalchemy_queries
+            except AttributeError:
+                queries = _app_ctx_stack.top.sqlalchemy_queries = []
 
-            g.sqlalchemy_queries.append(_DebugQueryTuple((
+            queries.append(_DebugQueryTuple((
                 statement, parameters, context._query_start_time, _timer(),
                 _calling_context(self.app_package)
             )))
@@ -298,7 +298,7 @@ def get_debug_queries():
         query was issued.  The exact format is undefined so don't try
         to reconstruct filename or function name.
     """
-    return g.get('sqlalchemy_queries', ())
+    return getattr(_app_ctx_stack.top, 'sqlalchemy_queries', [])
 
 
 class Pagination(object):
@@ -427,19 +427,23 @@ class BaseQuery(orm.Query):
             abort(404)
         return rv
 
-    def paginate(self, page=None, per_page=None, error_out=True):
+    def paginate(self, page=None, per_page=None, error_out=True, max_per_page=None):
         """Returns ``per_page`` items from page ``page``.
 
-        If no items are found and ``page`` is greater than 1, or if page is
-        less than 1, it aborts with 404.
-        This behavior can be disabled by passing ``error_out=False``.
-
         If ``page`` or ``per_page`` are ``None``, they will be retrieved from
-        the request query.
-        If the values are not ints and ``error_out`` is ``True``, it aborts
-        with 404.
-        If there is no request or they aren't in the query, they default to 1
-        and 20 respectively.
+        the request query. If ``max_per_page`` is specified, ``per_page`` will
+        be limited to that value. If there is no request or they aren't in the
+        query, they default to 1 and 20 respectively.
+
+        When ``error_out`` is ``True`` (default), the following rules will
+        cause a 404 response:
+
+        * No items are found and ``page`` is not 1.
+        * ``page`` is less than 1, or ``per_page`` is negative.
+        * ``page`` or ``per_page`` are not ints.
+
+        When ``error_out`` is ``False``, ``page`` and ``per_page`` default to
+        1 and 20 respectively.
 
         Returns a :class:`Pagination` object.
         """
@@ -469,8 +473,20 @@ class BaseQuery(orm.Query):
             if per_page is None:
                 per_page = 20
 
-        if error_out and page < 1:
-            abort(404)
+        if max_per_page is not None:
+            per_page = min(per_page, max_per_page)
+
+        if page < 1:
+            if error_out:
+                abort(404)
+            else:
+                page = 1
+
+        if per_page < 0:
+            if error_out:
+                abort(404)
+            else:
+                per_page = 20
 
         items = self.limit(per_page).offset((page - 1) * per_page).all()
 
@@ -548,63 +564,6 @@ class _EngineConnector(object):
             return rv
 
 
-def _should_set_tablename(cls):
-    """Traverse the model's MRO. If a primary key column is found before a
-    table or tablename, then a new tablename should be generated.
-
-    This supports:
-
-    * Joined table inheritance without explicitly naming sub-models.
-    * Single table inheritance.
-    * Inheriting from mixins or abstract models.
-
-    :param cls: model to check
-    :return: True if tablename should be set
-    """
-
-    for base in cls.__mro__:
-        d = base.__dict__
-
-        if '__tablename__' in d or '__table__' in d:
-            return False
-
-        for name, obj in iteritems(d):
-            if isinstance(obj, declared_attr):
-                obj = getattr(cls, name)
-
-            if isinstance(obj, sqlalchemy.Column) and obj.primary_key:
-                return True
-
-
-def camel_to_snake_case(name):
-    def _join(match):
-        word = match.group()
-
-        if len(word) > 1:
-            return ('_%s_%s' % (word[:-1], word[-1])).lower()
-
-        return '_' + word.lower()
-
-    return _camelcase_re.sub(_join, name).lstrip('_')
-
-
-class _BoundDeclarativeMeta(DeclarativeMeta):
-    def __new__(cls, name, bases, d):
-        # if tablename is set explicitly, move it to the cache attribute so
-        # that future subclasses still have auto behavior
-        if '__tablename__' in d:
-            d['_cached_tablename'] = d.pop('__tablename__')
-
-        return DeclarativeMeta.__new__(cls, name, bases, d)
-
-    def __init__(self, name, bases, d):
-        bind_key = d.pop('__bind_key__', None) or getattr(self, '__bind_key__', None)
-        DeclarativeMeta.__init__(self, name, bases, d)
-
-        if bind_key is not None and hasattr(self, '__table__'):
-            self.__table__.info['bind_key'] = bind_key
-
-
 def get_state(app):
     """Gets the state for the application"""
     assert 'sqlalchemy' in app.extensions, \
@@ -621,34 +580,6 @@ class _SQLAlchemyState(object):
         self.connectors = {}
 
 
-class Model(object):
-    """Base class for SQLAlchemy declarative base model.
-
-    To define models, subclass :attr:`db.Model <SQLAlchemy.Model>`, not this class.
-    To customize ``db.Model``, subclass this and pass it as ``model_class`` to :func:`SQLAlchemy`.
-    """
-
-    #: Query class used by :attr:`query`.
-    #: Defaults to :class:`SQLAlchemy.Query`, which defaults to :class:`BaseQuery`.
-    query_class = None
-
-    #: Convenience property to query the database for instances of this model using the current session.
-    #: Equivalent to ``db.session.query(Model)`` unless :attr:`query_class` has been changed.
-    query = None
-
-    _cached_tablename = None
-
-    @declared_attr
-    def __tablename__(cls):
-        if (
-            '_cached_tablename' not in cls.__dict__ and
-            _should_set_tablename(cls)
-        ):
-            cls._cached_tablename = camel_to_snake_case(cls.__name__)
-
-        return cls._cached_tablename
-
-
 class SQLAlchemy(object):
     """This class is used to control the SQLAlchemy integration to one
     or more Flask applications.  Depending on how you initialize the
@@ -797,16 +728,37 @@ class SQLAlchemy(object):
         return orm.sessionmaker(class_=SignallingSession, db=self, **options)
 
     def make_declarative_base(self, model, metadata=None):
-        """Creates the declarative base."""
-        base = declarative_base(cls=model, name='Model',
-                                metadata=metadata,
-                                metaclass=_BoundDeclarativeMeta)
+        """Creates the declarative base that all models will inherit from.
 
-        if not getattr(base, 'query_class', None):
-            base.query_class = self.Query
+        :param model: base model class (or a tuple of base classes) to pass
+            to :func:`~sqlalchemy.ext.declarative.declarative_base`. Or a class
+            returned from ``declarative_base``, in which case a new base class
+            is not created.
+        :param: metadata: :class:`~sqlalchemy.MetaData` instance to use, or
+            none to use SQLAlchemy's default.
 
-        base.query = _QueryProperty(self)
-        return base
+        .. versionchanged 2.3.0::
+            ``model`` can be an existing declarative base in order to support
+            complex customization such as changing the metaclass.
+        """
+        if not isinstance(model, DeclarativeMeta):
+            model = declarative_base(
+                cls=model,
+                name='Model',
+                metadata=metadata,
+                metaclass=DefaultMeta
+            )
+
+        # if user passed in a declarative base and a metaclass for some reason,
+        # make sure the base uses the metaclass
+        if metadata is not None and model.metadata is not metadata:
+            model.metadata = metadata
+
+        if not getattr(model, 'query_class', None):
+            model.query_class = self.Query
+
+        model.query = _QueryProperty(self)
+        return model
 
     def init_app(self, app):
         """This callback can be used to initialize an application for the
@@ -814,10 +766,13 @@ class SQLAlchemy(object):
         of an application not initialized that way or connections will
         leak.
         """
-        if 'SQLALCHEMY_DATABASE_URI' not in app.config:
+        if (
+            'SQLALCHEMY_DATABASE_URI' not in app.config and
+            'SQLALCHEMY_BINDS' not in app.config
+        ):
             warnings.warn(
-                'SQLALCHEMY_DATABASE_URI not set. Defaulting to '
-                '"sqlite:///:memory:".'
+                'Neither SQLALCHEMY_DATABASE_URI nor SQLALCHEMY_BINDS is set. '
+                'Defaulting SQLALCHEMY_DATABASE_URI to "sqlite:///:memory:".'
             )
 
         app.config.setdefault('SQLALCHEMY_DATABASE_URI', 'sqlite:///:memory:')
@@ -948,14 +903,15 @@ class SQLAlchemy(object):
             return reference_app
 
         if current_app:
-            return current_app
+            return current_app._get_current_object()
 
         if self.app is not None:
             return self.app
 
         raise RuntimeError(
-            'application not registered on db instance and no application'
-            'bound to current context'
+            'No application found. Either work inside a view function or push'
+            ' an application context. See'
+            ' http://flask-sqlalchemy.pocoo.org/contexts/.'
         )
 
     def get_tables_for_bind(self, bind=None):
@@ -1029,6 +985,15 @@ class SQLAlchemy(object):
         )
 
 
+class _BoundDeclarativeMeta(DefaultMeta):
+    def __init__(cls, name, bases, d):
+        warnings.warn(FSADeprecationWarning(
+            '"_BoundDeclarativeMeta" has been renamed to "DefaultMeta". The'
+            ' old name will be removed in 3.0.'
+        ), stacklevel=3)
+        super(_BoundDeclarativeMeta, cls).__init__(name, bases, d)
+
+
 class FSADeprecationWarning(DeprecationWarning):
     pass
 
